import tempfile

import unittest
from pathlib import Path

from ..cache import SimpleCache


class SimpleCacheTest(unittest.TestCase):
    def test_basic(self) -> None:
        with tempfile.TemporaryDirectory() as d:
            pd = Path(d)

            c = SimpleCache(pd)

            self.assertEqual(None, c.get("foo"))
            c.set("foo", b"value\n")
            self.assertEqual(b"value\n", c.get("foo"))

            self.assertEqual(
                pd
                / "0"
                / "b"
                / "e"
                / "e"
                / "c"
                / "0beec7b5ea3f0fdbc95d0dd47f3c5bc275da8a33",
                c._local_path("foo"),
            )
            self.assertEqual(6, c._local_path("foo").stat().st_size)

    def test_basic_suffix(self) -> None:
        with tempfile.TemporaryDirectory() as d:
            pd = Path(d)

            c = SimpleCache(pd, suffix="suf")

            self.assertEqual(None, c.get("foo"))
            c.set("foo", b"value\n")
            self.assertEqual(b"value\n", c.get("foo"))

            self.assertEqual(
                pd
                / "suf"
                / "0"
                / "b"
                / "e"
                / "e"
                / "c"
                / "0beec7b5ea3f0fdbc95d0dd47f3c5bc275da8a33",
                c._local_path("foo"),
            )
            self.assertEqual(6, c._local_path("foo").stat().st_size)

    def test_additional_coverage(self) -> None:
        # cover the appdirs call and ensure it is some kind of path we can
        # manipulate.
        SimpleCache()._local_path("foo")



import os
from hashlib import sha1
from pathlib import Path
from tempfile import mkstemp
from typing import Optional

import appdirs


class SimpleCache:
    """
    An extremely simple cache for storing immutable objects.
    """

    def __init__(
        self, cache_path: Optional[Path] = None, suffix: Optional[str] = None
    ) -> None:
        if not cache_path:
            cache_path = Path(appdirs.user_cache_dir("hdeps", "python-packaging"))
        if suffix:
            cache_path = cache_path / suffix

        self.cache_path = cache_path
        self.hash_factory = sha1
        self.stats = {"hits": 0, "pass": 0, "sets": 0}

    def _local_path(self, key: str) -> Path:
        h = self.hash_factory(key.encode("utf-8")).hexdigest()
        return self.cache_path.joinpath(*h[:5], h)

    def get(self, key: str) -> Optional[bytes]:
        p = self._local_path(key)
        try:
            data = p.read_bytes()
            self.stats["hits"] += 1
            return data
        except OSError:
            self.stats["pass"] += 1
            return None

    def set(self, key: str, value: bytes) -> None:
        p = self._local_path(key)
        p.parent.mkdir(parents=True, exist_ok=True)
        (fd, temp_name) = mkstemp(f".{os.getpid()}", prefix=p.name, dir=p.parent)
        with os.fdopen(fd, "wb") as f:
            f.write(value)

        os.replace(temp_name, p)
        self.stats["sets"] += 1


class NoCache(SimpleCache):
    def get(self, key: str) -> Optional[bytes]:
        return None

    def set(self, key: str, value: bytes) -> None:
        pass



from __future__ import annotations

import logging
import sys
import tempfile

from collections import defaultdict
from dataclasses import dataclass, field
from email import message_from_string
from functools import partial
from pathlib import Path
from tarfile import TarFile
from typing import Dict, Iterable, List, Optional, Sequence, Set, Tuple, TypeVar, Union
from zipfile import ZipFile

import metadata_please
from indexurl import get_index_url

from keke import kev
from packaging.requirements import InvalidRequirement, Requirement
from packaging.utils import canonicalize_name
from packaging.version import InvalidVersion, Version
from pypi_simple import DistributionPackage, ProjectPage, PyPISimple

from requests.sessions import Session

from seekablehttpfile import SeekableHttpFile
from seekablehttpfile.core import get_range_requests

from .cache import SimpleCache
from .types import CanonicalName

LOG = logging.getLogger(__name__)

T = TypeVar("T")
D = TypeVar("D")


def first(it: Iterable[T], default: D) -> Union[T, D]:
    """like any(...) but returns the first truthy value"""
    for x in it:
        if x:
            return x
    return default


@dataclass(frozen=True)
class Project:
    name: CanonicalName
    versions: Dict[Version, ProjectVersion]

    @classmethod
    def from_pypi_simple_project_page(cls, project_page: ProjectPage) -> Project:
        vers: Dict[Version, List[DistributionPackage]] = defaultdict(list)
        for dp in project_page.packages:
            if dp.version is None:
                LOG.debug("Ignore unset version in %s", dp.filename)
                continue
            try:
                vers[Version(dp.version)].append(dp)
            except InvalidVersion:
                LOG.debug("Ignore invalid version %s in %s", dp.version, dp.filename)
        return cls(
            name=CanonicalName(canonicalize_name(project_page.project)),
            versions={
                v: ProjectVersion(v, tuple(pkgs)) for v, pkgs in sorted(vers.items())
            },
        )


def filter_requires_txt_names(names: List[str]) -> List[str]:
    # TODO: I can't remember why it's <= 2 rather than a specific count
    return [
        name
        for name in names
        if name.endswith("/requires.txt") and name.count("/") <= 2
    ]


@dataclass(frozen=True)
class ProjectVersion:
    version: Version
    packages: Tuple[DistributionPackage, ...] = field(hash=False)

    @property
    def requires_python(self) -> Optional[str]:
        return first((dp.requires_python for dp in self.packages), None)

    @property
    def yanked(self) -> bool:
        return first((dp.is_yanked for dp in self.packages), False)

    @property
    def has_sdist(self) -> bool:
        return any(dp.package_type == "sdist" for dp in self.packages)

    @property
    def has_wheel(self) -> bool:
        return any(dp.package_type == "wheel" for dp in self.packages)

    def get_deps(
        self, ps: PyPISimple, session: Session, extracted_metadata_cache: SimpleCache
    ) -> BasicMetadata:
        best_pkg: Optional[DistributionPackage] = None
        best_score = 0
        with kev("score", count=len(self.packages)):
            for pkg in self.packages:
                if pkg.has_metadata:
                    score = 100
                elif pkg.package_type == "wheel":
                    score = 90
                elif pkg.package_type == "sdist" and pkg.filename.endswith(".zip"):
                    score = 50
                elif pkg.package_type == "sdist":
                    score = 30
                else:
                    LOG.debug("Cannot load metadata from %r", pkg.filename)
                    continue
                assert pkg
                if score > best_score:
                    best_pkg = pkg
                    best_score = score
                elif score == best_score and pkg.filename < best_pkg.filename:  # type: ignore[union-attr]
                    # Ensure consistent ordering for good cache hits
                    best_pkg = pkg

        if best_pkg is None:
            LOG.warning(
                "Cannot load metadata from version %s, no matching packages in %s",
                self.version,
                [dp.filename for dp in self.packages],
            )
            return BasicMetadata()

        md: str
        # Ensure we don't accidentally use this later on; this was the source of a pesky cache bug
        del pkg
        if best_pkg.has_metadata:
            if md_bytes := extracted_metadata_cache.get(best_pkg.url):
                md = md_bytes.decode("utf-8")
            else:
                with kev("pypi_simple.get_package_metadata"):
                    # pypi-simple does the decode for you today; this wastes a
                    # ton of time trying to guess charset, until
                    # https://github.com/jwodder/pypi-simple/pull/22 is merged.
                    md = ps.get_package_metadata(best_pkg)
                    extracted_metadata_cache.set(best_pkg.url, md.encode("utf-8"))
        elif best_pkg.package_type == "wheel":
            # Wheels can be loaded incrementally, but also provide richer, more
            # reliable metadata.
            if md_bytes := extracted_metadata_cache.get(best_pkg.url):
                md = md_bytes.decode("utf-8")
            else:
                with kev("extract metadata remote", url=best_pkg.url):
                    f = SeekableHttpFile(
                        best_pkg.url,
                        get_range=partial(get_range_requests, session=session),
                        check_etag=False,
                    )
                    zf = ZipFile(f)  # type: ignore[arg-type,call-overload,unused-ignore]

                    # These two lines come from warehouse itself
                    name, version, _ = best_pkg.filename.split("-", 2)
                    md_bytes = metadata_please.from_wheel(zf, name)
                    extracted_metadata_cache.set(best_pkg.url, md_bytes)
                md = md_bytes.decode("utf-8")
        elif best_pkg.package_type == "sdist":
            key = best_pkg.url + "#requires.txt"
            if (md_bytes := extracted_metadata_cache.get(key)) is not None:
                md = md_bytes.decode("utf-8")
            else:
                if best_pkg.filename.endswith(".zip"):
                    with kev("extract sdist metadata remote", url=best_pkg.url):
                        f = SeekableHttpFile(
                            best_pkg.url,
                            get_range=partial(get_range_requests, session=session),
                            check_etag=False,
                        )
                        zf = ZipFile(f)  # type: ignore[arg-type,call-overload,unused-ignore]
                        md_bytes = metadata_please.from_zip_sdist(zf)
                else:
                    with kev("extract tar metadata", url=best_pkg.url):
                        with tempfile.TemporaryDirectory() as d:
                            local_path = Path(d, best_pkg.filename)
                            ps.download_package(
                                best_pkg, path=local_path, verify=bool(best_pkg.digests)
                            )
                            with TarFile.open(local_path) as tf:
                                md_bytes = metadata_please.from_tar_sdist(tf)

                extracted_metadata_cache.set(key, md_bytes)
                md = md_bytes.decode("utf-8")

        else:
            raise NotImplementedError(best_pkg)

        reqs: List[Requirement] = []
        extras: List[str] = []
        if md is not None:
            with kev("transform"):
                msg = message_from_string(md)
                if t := msg.get_all("Requires-Dist"):
                    for dep in t:
                        try:
                            reqs.append(Requirement(dep))
                        except InvalidRequirement:
                            LOG.warning(
                                "Skipping invalid requirement %r",
                                dep,
                            )

                if t := msg.get_all("Provides-Extra"):
                    extras = t
        return BasicMetadata(
            reqs, extras, has_sdist=self.has_sdist, has_wheel=self.has_wheel
        )


def convert_sdist_requires(data: str) -> Tuple[List[str], Set[str]]:
    # This is reverse engineered from looking at a couple examples, but there
    # does not appear to be a formal spec.  Mentioned at
    # https://setuptools.readthedocs.io/en/latest/formats.html#requires-txt
    current_markers = None
    extras: Set[str] = set()
    lst: List[str] = []
    for line in data.splitlines():
        line = line.strip()
        if not line:
            continue
        elif line[:1] == "[" and line[-1:] == "]":
            current_markers = line[1:-1]
            if ":" in current_markers:
                # absl-py==0.9.0 and requests==2.22.0 are good examples of this
                extra, markers = current_markers.split(":", 1)
                if extra:
                    extras.add(extra)
                    current_markers = f"({markers}) and extra == {extra!r}"
                else:
                    current_markers = markers
            else:
                # this is an extras_require
                current_markers = f"extra == {current_markers!r}"
        else:
            if current_markers:
                lst.append(f"{line}; {current_markers}")
            else:
                lst.append(line)
    return lst, extras


@dataclass
class BasicMetadata:
    reqs: Sequence[Requirement] = ()
    extras: Sequence[str] = ()
    has_sdist: bool = False
    has_wheel: bool = False


if __name__ == "__main__":  # pragma: no cover
    from .session import get_retry_session

    s = get_retry_session()
    x = PyPISimple(get_index_url(), session=s)
    p = Project.from_pypi_simple_project_page(
        x.get_project_page(CanonicalName(sys.argv[1]))
    )
    if len(sys.argv) > 2:
        ver = Version(sys.argv[2])
    else:
        ver = list(p.versions)[-1]
    v = p.versions[ver]
    print(f"{sys.argv[1]}=={ver}", v.get_deps(x, s, SimpleCache()))



import io
import unittest
from pathlib import Path
from unittest.mock import patch

from click.testing import CliRunner

from packaging.requirements import Requirement
from pypi_simple import PyPISimple

from ..cache import NoCache
from ..markers import EnvironmentMarkers
from ..resolution import Walker

from ._fake_session import FakeSession


class ResolutionTest(unittest.TestCase):
    # These two tests are just belt-and-suspenders; they do are identical to two of the
    # cli_scenarios tests but without invoking through click, just to ensure the api
    # remains stable.
    def test_simple_integration(self) -> None:
        session = FakeSession(Path(__file__).parent / "fixtures")
        pypi_simple = PyPISimple(session=session)  # type: ignore[arg-type]
        env_markers = EnvironmentMarkers.from_args("3.7.5", None)
        runner = CliRunner()
        with runner.isolated_filesystem():
            walker = Walker(
                1,
                env_markers,
                pypi_simple,
                session,  # type: ignore[arg-type]
                extracted_metadata_cache=NoCache(),
            )
            walker.feed(Requirement("batman==1"))
            walker.drain()

        new_stdout = io.StringIO()
        with patch("sys.stdout", new_stdout):
            walker.print_flat()

        self.assertEqual(
            """\
robin==1.0
batman==1.0
""",
            new_stdout.getvalue(),
        )

        new_stdout = io.StringIO()
        with patch("sys.stdout", new_stdout):
            walker.print_tree()

        self.assertEqual(
            """\
batman (==1.0) via ==1
. robin (==1.0) via ==1.0
""",
            new_stdout.getvalue(),
        )

    def test_reuse_integration(self) -> None:
        session = FakeSession(Path(__file__).parent / "fixtures")
        pypi_simple = PyPISimple(session=session)  # type: ignore[arg-type]
        env_markers = EnvironmentMarkers.from_args("3.7.5", None)
        runner = CliRunner()
        with runner.isolated_filesystem():
            walker = Walker(
                1,
                env_markers,
                pypi_simple,
                session,  # type: ignore[arg-type]
                extracted_metadata_cache=NoCache(),
            )
            walker.feed(Requirement("batman==1"))
            walker.drain()

        new_stdout = io.StringIO()
        with patch("sys.stdout", new_stdout):
            walker.print_flat()

        self.assertEqual(
            """\
robin==1.0
batman==1.0
""",
            new_stdout.getvalue(),
        )

        new_stdout = io.StringIO()
        with patch("sys.stdout", new_stdout):
            walker.print_tree()

        self.assertEqual(
            """\
batman (==1.0) via ==1
. robin (==1.0) via ==1.0
""",
            new_stdout.getvalue(),
        )



from .cache import SimpleCacheTest
from .cli_scenarios import CliScenariosTest
from .markers import EnvironmentMarkersTest
from .projects import ProjectMetadataTest
from .requirements import RequirementsTest
from .resolution import ResolutionTest

# from .session import LiveSessionTest

__all__ = [
    "CliScenariosTest",
    "EnvironmentMarkersTest",
    "SimpleCacheTest",
    "RequirementsTest",
    "ResolutionTest",
    "ProjectMetadataTest",
    # "LiveSessionTest",
]



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Callable, List, NewType, Optional, Tuple

from packaging.markers import Marker
from packaging.specifiers import SpecifierSet
from packaging.version import Version

CanonicalName = NewType("CanonicalName", str)
VersionCallback = Callable[[CanonicalName], Optional[str]]

ChoiceKeyType = Tuple[CanonicalName, Version, Tuple[str, ...]]


@dataclass
class Choice:
    project: CanonicalName = field(repr=True)
    version: Version = field(repr=True)
    # TODO extras should be normalized, thus CanonicalName
    extras: Tuple[str, ...] = field(default_factory=tuple)
    deps: List[Edge] = field(default_factory=list)
    has_sdist: bool = False
    has_wheel: bool = False

    def key(self) -> ChoiceKeyType:
        return (self.project, self.version, self.extras)


@dataclass
class Edge:
    target: Choice = field(repr=True)
    specifier: Optional[SpecifierSet]
    markers: Optional[Marker]
    note: Optional[str]



import logging
from typing import Dict, List, Optional, Tuple

from keke import kev

from packaging.requirements import Requirement
from packaging.specifiers import InvalidSpecifier, SpecifierSet
from packaging.version import Version

from .markers import EnvironmentMarkers
from .projects import Project
from .types import VersionCallback

LOG = logging.getLogger(__name__)


def find_best_compatible_version(
    project: Project,
    req: Requirement,
    env_markers: EnvironmentMarkers,
    already_chosen: Optional[Version],
    current_version_callback: VersionCallback,
) -> Version:
    # Handle requires_python first, so we can produce a better error message
    # when there are no version-compatible candidates (before we even get to
    # req.specifier)
    python_version_str = env_markers.python_full_version
    assert python_version_str is not None
    python_version = Version(python_version_str)

    requires_python_cache: Dict[str, bool] = {}

    def requires_python_match(v: Version) -> bool:
        pv = project.versions[v]
        if not pv.requires_python:
            return True
        if verdict := requires_python_cache.get(pv.requires_python) is not None:
            return verdict
        try:
            specifier_set = SpecifierSet(pv.requires_python)
            verdict = python_version in specifier_set
        except InvalidSpecifier as e:
            LOG.debug(
                "Ignore %s==%s has invalid requires_python %r but including anway",
                project.name,
                v,
                e,
            )
            verdict = True
        requires_python_cache[pv.requires_python] = verdict
        return verdict

    possible: List[Version] = []

    specifier_matched = False
    with kev("reverse"):
        rev = reversed(project.versions.keys())
    with kev("initial filter"):
        for version in req.specifier.filter(rev):
            specifier_matched = True
            if requires_python_match(version):
                # Only keep the first one
                possible.append(version)
                break

    # If the current version is a non-public version, we need to add it back
    # here.
    with kev("current_version_callback"):
        cur = current_version_callback(project.name)
    cur_v: Optional[Version] = None
    if cur:
        cur_v = Version(cur)
        # Here be dragons: if we already filtered the current version out for
        # requires_python, then don't bother adding it back in; this is only
        # intended for non-public version reuse.
        if cur_v not in project.versions:
            possible.append(cur_v)

    if already_chosen:
        possible.append(already_chosen)

    if not possible:
        if specifier_matched:
            raise ValueError(
                f"{project.name} has no {python_version}-compatible release"
            )
        else:
            raise ValueError(f"{project.name} has no release matching {req.specifier}")

    # The documentation for SepcifierSet.filter notes that it handles the logic
    # for whether to include prereleases, so we don't need that here.
    LOG.debug("possible %s", possible)
    with kev("filter by specifier"):
        # This should only ever be ~3 items now!
        possible = list(req.specifier.filter(possible))
    if not possible:
        if not list(req.specifier.filter(project.versions.keys())):
            # Referencing the dragon above, if we had a current version and it was
            # unsuitable, then we still output a generic message.  Note that > does not
            # set the prerelease bit, but >= does.
            raise ValueError(
                f"{project.name} has no release with constraint {req.specifier}"
            )
        else:
            raise ValueError(
                f"{project.name} has no {python_version}-compatible release with constraint {req.specifier}"
            )

    # TODO: yanked support

    # The sort key is
    # ( matches_already_chosen: bool,
    #   matches_current_version: bool,
    #   recency_index: int,
    #   version: Version )
    # so that after sorting the last item is the "best" one.

    with kev("final sort"):
        xform_possible: List[Tuple[bool, bool, int, Version]] = sorted(
            (p == already_chosen, p == cur_v, i, p) for (i, p) in enumerate(possible)
        )

    return xform_possible[-1][3]
